#!/bin/python
import os
import time

from vagd import Vagd, Qegd, Shgd, Dogd, wrapper, Box
from pwn import *

GDB_OFF = 0x555555555000
IP = ''
PORT = 0
BINARY = './bin/sysinfo'
ARGS = []
ENV = {}
API = True
GDB = f"""
b main
c"""

context.binary = exe = ELF(BINARY, checksec=False)
context.aslr = False

byt = lambda x: str(x).encode()


def vms():
    vm = Vagd(exe.path, vbox=Box.UBUNTU_FOCAL64, tmp=True, fast=True, ex=True)
    yield vm
    vm._v.halt()
    vm = Dogd(exe.path, user='ubuntu', image=Box.DOCKER_FOCAL, tmp=True, ex=True, fast=True)
    yield vm
    vm._client.containers.get(vm._id).kill()
    vm = Qegd(exe.path, user='ubuntu', img=Box.CLOUDIMAGE_FOCAL, tmp=True, ex=True, fast=True)
    yield vm
    yield Shgd(exe.path, user=vm._user, port=vm._port, tmp=True, ex=True, fast=True)
    os.system('kill $(pgrep qemu)')
    yield wrapper.Empty()


for vm in vms():
    t = vm.start(argv=ARGS, env=ENV, gdbscript=GDB, api=API)

    g = wrapper.GDB(t)
    g.execute('p "PWN"')
    g.execute('c')

    t.recvall()

print("Everything executed without errors")



import os
import pwn
from vagd.virts.pwngd import Pwngd

_GENERATE_KEYPAIR = 'ssh-keygen -q -t ed25519 -f {keyfile} -N ""'


def generate_keypair():
    """
    generate a keypair in .vagd directory
    """
    if not (os.path.exists(Pwngd.KEYFILE) and os.path.exists(Pwngd.KEYFILE + '.pub')):
        pwn.log.info("No Keypair was found. Generating new keypair")
        os.system(_GENERATE_KEYPAIR.format(keyfile=Pwngd.KEYFILE))


def is_port_in_use(port: int) -> bool:
    """
    check if a port is currently used
    :param port: port to check
    :return: if the port is already used
    """
    import socket
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0


def first_free_port(start: int = 2222, tries: int = 101) -> int:
    """
    returns first free port starting from start (max increments = tries)
    :param start: start of port search
    :param tries: number of tries to increment ports
    :return: first free port
    """
    for i in range(tries):
        port = start + i
        if not is_port_in_use(port):
            return port

    pwn.log.error(f'No free port inside range {start}-{start+tries}')



import pwn
from vagd.virts.pwngd import Pwngd
class Shgd(Pwngd):
    """ ssh interface for pwntools """

    DEFAULT_HOST = 'localhost'
    DEFAULT_PORT = 22
    DEFAULT_USER = 'root'

    _user: str
    _host: str
    _port: int
    _keyfile: str

    def _vm_setup(self) -> None:
        """
        pass
        """
        pass

    def _ssh_setup(self) -> None:
        """
        setups ssh connection to remote
        """
        self._ssh = pwn.ssh(
            user=self._user,
            host=self._host,
            port=self._port,
            keyfile=self._keyfile,
            ignore_config=True
        )

    def __init__(self,
                 binary: str,
                 user: str = DEFAULT_USER,
                 host: str = DEFAULT_HOST,
                 port: int = DEFAULT_PORT,
                 keyfile: str = Pwngd.KEYFILE,
                 **kwargs):
        """

        :param binary: binary to execute
        :param user: ssh user
        :param host: ssh hostname
        :param port: ssh port
        :param keyfile: ssh keyfile (default in .vagd)
        :param kwargs: parameters to pass through to super
        """
        self._user = user
        self._host = host
        self._port = port
        self._keyfile = keyfile

        self._ssh_setup()

        super().__init__(binary=binary, **kwargs)





import os
import pwn
import docker

from vagd import templates, helper
from vagd.box import Box
from vagd.virts.shgd import Shgd
from vagd.virts.pwngd import Pwngd


class Dogd(Shgd):
    """
    | Docker virtualization for pwntools
    | SSH from cmd
    .. code-block:: bash

        ssh -o "StrictHostKeyChecking=no" -i .vagd/keyfile -p $(cut .vagd/docker.lock -d":" -f 2) vagd@0.0.0.0

    | connect with docker exec
    .. code-block:: bash

       docker exec -it $(cut ./.vagd/docker.lock -d":" -f 1) /bin/bash

    | Kill from cmd:
    .. code-block:: bash

        docker kill $(cut ./.vagd/docker.lock -d":" -f 1)

    | Docker containers are automatically removed after they stop
    | Docker images need to be manually removed from docker
    .. code-block:: bash

        docker images # list images
        docker rmi <id> # remove correct image
    """

    _image: str
    _user: str
    _port: int
    _client: docker.client
    _id: str

    DEFAULT_USER = 'vagd'
    DEFAULT_PORT = 2222
    DEFAULT_IMAGE = Box.DOCKER_FOCAL

    DEFAULT_PACKAGES = Pwngd.DEFAULT_PACKAGES + ["openssh-server"]
    DEFAULT_DOCKERFILE = Pwngd.LOCAL_DIR + "Dockerfile"
    LOCKFILE = Pwngd.LOCAL_DIR + 'docker.lock'

    def _create_dockerfile(self):

        pwn.log.info(f'create new Dockerfile at f{self._dockerfile}')
        if not os.path.exists(Pwngd.KEYFILE):
            helper.generate_keypair()

        with open(Dogd.DEFAULT_DOCKERFILE, 'w') as dockerfile:
            dockerfile.write(
                templates.DOCKER_TEMPLATE.format(image=self._image,
                                                 packages=' '.join(Dogd.DEFAULT_PACKAGES),
                                                 user=self._user,
                                                 keyfile=os.path.basename(Pwngd.KEYFILE + '.pub')))

    def _create_docker_instance(self):
        pwn.log.info('starting docker instance')
        self._port = helper.first_free_port(Dogd.DEFAULT_PORT)
        container = self._client.containers.run(self._bimage, ports={'22/tcp': self._port}, detach=True, remove=True)
        self._id = container.id
        pwn.log.info(f'started docker instance {container.short_id}')
        with open(Dogd.LOCKFILE, 'w') as lockfile:
            lockfile.write(container.id + ':' + str(self._port))

    def _build_image(self):
        pwn.log.info('building docker image')
        return self._client.images.build(path=os.path.dirname(self._dockerfile))[0]

    def _vm_create(self):

        if self._dockerfile == Dogd.DEFAULT_DOCKERFILE and not os.path.exists(self._dockerfile):
            self._create_dockerfile()

        self._bimage = self._build_image()

        self._create_docker_instance()

    def _vm_setup(self) -> None:
        self._client = docker.from_env()
        if not os.path.exists(Dogd.LOCKFILE):
            pwn.log.info(f'No Lockfile {Dogd.LOCKFILE} found, creating new Docker Instance')
            self._vm_create()
        else:
            with open(Dogd.LOCKFILE, 'r') as lockfile:
                data = lockfile.readline().split(':')
                self._id = data[0]
                self._port = int(data[1])
            if not helper.is_port_in_use(self._port):
                pwn.log.info(f'Lockfile {Dogd.LOCKFILE} found, port not used, creating new container')
                self._vm_create()
            else:
                pwn.log.info(
                    f'Lockfile {Dogd.LOCKFILE} found, Docker Instance f{self._client.containers.get(self._id).short_id}')

    def __init__(self,
                 binary: str,
                 image: str = DEFAULT_IMAGE,
                 user: str = DEFAULT_USER,
                 **kwargs):
        """

        :param binary: binary to execute
        :param image: docker base image
        :param user: name of user on docker container
        :param kwargs: parameters to pass through to super
        """
        self._image = image
        self._dockerfile = Dogd.DEFAULT_DOCKERFILE
        self._user = user

        self._vm_setup()

        super().__init__(binary=binary,
                         user=self._user,
                         port=self._port,
                         **kwargs)



import os
import re
import pwn
import vagrant
import fileinput
from shutil import which

from vagd import templates
from vagd.box import Box
from vagd.virts.pwngd import Pwngd
from vagd.virts.shgd import Shgd


class Vagd(Shgd):
    """
    | Vagrant Virtualization for pwntools
    |
    | SSH from cmd:
    .. code-block::  bash

        VAGRANT_CWD=.vagd vagrant ssh

    | halt from cmd
    .. code-block:: bash

        VAGRANT_CWD=.vagd vagrant halt

    | destroy from cmd
    .. code-block:: bash

        VAGRANT_CWD=.vagd vagrant destroy
    """

    VAGRANTFILE_PATH = Pwngd.LOCAL_DIR + 'Vagrantfile'
    VAGRANTFILE_BOX = 'config.vm.box'
    VAGRANT_BOX = Box.UBUNTU_FOCAL64

    _box: str
    _vagrantfile: str
    _v: vagrant

    def _get_box(self) -> str:
        """
        returns box of current vagrantfile
        @:rtype box name of Vagrantfile
        """
        with open(self._vagrantfile, 'r') as vagrantfile:
            for line in vagrantfile.readlines():
                if Vagd.VAGRANTFILE_BOX in line:
                    pattern = fr'{Vagd.VAGRANTFILE_BOX} = "(.*?)"'
                    match = re.search(pattern, line, re.DOTALL)
                    if match:
                        return match.group(1)
        return ''

    def _vm_setup(self) -> None:
        """
        setup vagrant machine creates new one if no Vagrantfile is specified or box does not match
        """

        if not os.path.isfile(self._vagrantfile):
            pwn.log.info('creating new Vagrantfile')
            vagrant_config = templates.VAGRANT_TEMPLATE.format(box=self._box, packages=' '.join(Pwngd.DEFAULT_PACKAGES))
            with open(self._vagrantfile, 'w') as file:
                file.write(vagrant_config)
            pwn.log.info('initialing new vagrant vm might take a while')
            self._v.up()

        elif self._get_box() != self._box:
            pwn.log.info('new box detected destroying old machine')
            self._v.destroy()
            for line in fileinput.input(self._vagrantfile, inplace=True):
                if Vagd.VAGRANTFILE_BOX in line:
                    line = f'{Vagd.VAGRANTFILE_BOX} = "{self._box}"\n'
                print(line, end='')
            pwn.log.info('initialing new vagrant vm might take a while')
            self._v.up()

        if self._v.status()[0].state != 'running':
            pwn.log.info('starting existing vagrant machine')
            self._v.up()

    def __init__(self,
                 binary: str,
                 vagrantfile: str = VAGRANTFILE_PATH,
                 vbox: str = None,
                 **kwargs):
        """

        :param binary: binary for VM debugging
        :param vbox: vagrant box to use
        :param vagrantfile: location of Vagrantfile
        :param kwargs: arguments to pass through to super
        """

        if not which('vagrant'):
            pwn.log.error('vagrant isn\'t installed')

        if not os.path.exists(Pwngd.LOCAL_DIR):
            os.makedirs(Pwngd.LOCAL_DIR)

        self._vagrantfile = vagrantfile

        if vbox is None:
            if os.path.exists(vagrantfile):
                vbox = self._get_box()
            else:
                vbox = Vagd.VAGRANT_BOX

        self._box = vbox
        self._v = vagrant.Vagrant(os.path.dirname(vagrantfile))

        self._vm_setup()

        super().__init__(binary=binary,
                         user=self._v.user(),
                         host=self._v.hostname(),
                         port=int(self._v.port()),
                         keyfile=self._v.keyfile(), **kwargs)



import os
import sys

import pwn
import time
import requests

from typing import Dict
from urllib.parse import urlparse
from shutil import which, copyfile

from vagd import helper
from vagd.box import Box
from vagd.virts.pwngd import Pwngd


class Qegd(Pwngd):
    """
    | QEMU Virtualization for pwntools
    | SSH from cmd

    .. code-block:: bash

        ssh -o "StrictHostKeyChecking=no" -i .vagd/keyfile -p $(cat .vagd/qemu.lock) ubuntu@0.0.0.0

    | Kill from cmd:

    .. code-block:: bash

        kill $(pgrep qemu)

    | Qemu images are cached in the home directory: :code:`~/.vagd/qemu-imgs/`
    |
    | current used images are stored in the local directory: :code:`./.vagd/current.img`
    | These should be deleted automatically, but if a machine gets improperly stopped
    | (shutdown host while vm is running) it might remain and use up space. You can find remaining images with:

    .. code-block:: bash
        
        find ~/ -name current.img
        rm <path/current.img>

    """

    DEFAULT_IMG = Box.CLOUDIMAGE_FOCAL
    QEMU_DIR = Pwngd.LOCAL_DIR
    IMGS_DIR = Pwngd.HOME_DIR + 'qemu-imgs/'
    DEFAULT_USER = 'ubuntu'
    DEFAULT_HOST = '0.0.0.0'
    DEFAULT_PORT = 2222
    ARM_FLASH = "/usr/share/AAVMF/AAVMF_CODE.fd"
    DEFAULT_QEMU_ARM_PFLASH_OPTIONS = ""
    DEFAULT_QEMU_ARM_PFLASH = QEMU_DIR + "flash.img"
    DEFAULT_QEMU_CMD = "qemu-system-x86_64"
    DEFAULT_QEMU_ARM_CMD = "qemu-system-aarch64"
    DEFAULT_QEMU_MACHINE_PREFIX = "-machine"
    DEFAULT_QEMU_MACHINE = "accel=kvm,type=q35"
    DEFAULT_QEMU_ARM_MACHINE = "virt"
    DEFAULT_QEMU_CPU_PREFIX = "-cpu"
    DEFAULT_QEMU_CPU = "host"
    DEFAULT_QEMU_ARM_CPU = "cortex-a57"
    DEFAULT_QEMU_PFLASH_PREFIX = "-pflash"

    _img: str
    _new: bool = False
    _local_img: str
    _user: str
    _host: str
    _port: int
    _ports: Dict[int, int]
    _qemu: str
    _cpu: str
    _machine: str
    _keyfile: str

    @staticmethod
    def _is_local(url) -> bool:
        """
        check if provided url is local or remote
        :param url: url to check
        :return: if the url is local or remote
        """
        url_parsed = urlparse(url)
        if url_parsed.scheme in ('file', ''):  # Possibly a local file
            return os.path.exists(url_parsed.path)
        return False

    CURRENT_IMG = QEMU_DIR + "current.img"

    def _set_local_img(self):
        """
        get local image for qemu machine
        """
        if Qegd._is_local(self._img):
            pwn.log.info("Using local image")
            self._local_img = self._img
        else:
            if not os.path.exists(Qegd.IMGS_DIR):
                os.makedirs(Qegd.IMGS_DIR)
            self._local_img = Qegd.IMGS_DIR + urlparse(self._img).path.rsplit('/', 1)[-1]
            if not os.path.exists(self._local_img):
                pwn.log.info("online qemu image starting download")
                img = requests.get(self._img)
                with open(self._local_img, 'wb') as imgfile:
                    pwn.log.info(f"saving image to {self._local_img}")
                    imgfile.write(img.content)
        copyfile(self._local_img, Qegd.CURRENT_IMG)

    METADATA_FILE = QEMU_DIR + 'metadata.yaml'
    _METADATA = """instance-id: iid-local01
local-hostname: cloudimg
"""
    USER_DATA_FILE = QEMU_DIR + 'user-data.yaml'
    _USER_DATA = """#cloud-config
ssh_authorized_keys:
  - {pubkey}
"""

    SEED_FILE = QEMU_DIR + "seed.img"

    _GENERATE_SEED_IMG = f'cloud-localds {SEED_FILE} {USER_DATA_FILE} {METADATA_FILE}'

    def _setup_seed(self):
        """
        create seed.img with config data like ssh keypair in .qemu
        """
        if not which('cloud-localds'):
            pwn.log.error("cloud-image-utils is not installed")
        if not os.path.exists(Qegd.SEED_FILE):
            pwn.log.info(f"{Qegd.SEED_FILE} not found generating new one")
            if not os.path.exists(Qegd.METADATA_FILE):
                pwn.log.info(f"{Qegd.METADATA_FILE} not found generating new one")
                with open(Qegd.METADATA_FILE, 'w') as metadata_file:
                    metadata_file.write(Qegd._METADATA)
            if not os.path.exists(Qegd.USER_DATA_FILE):
                pwn.log.info(f"{Qegd.USER_DATA_FILE} not found generating new one")
                helper.generate_keypair()
                with open(Qegd.USER_DATA_FILE, 'w') as user_data_file:
                    with open(Pwngd.KEYFILE + '.pub', 'r') as pubkey_file:
                        pubkey = pubkey_file.readline()
                    user_data_file.write(Qegd._USER_DATA.format(pubkey=pubkey))
            os.system(Qegd._GENERATE_SEED_IMG)

    _QEMU_PORT_FORWARDING = ',hostfwd=tcp::{host}-:{guest}'
    _QEMU_START = "{qemu} " \
                  + "{machine} " \
                  + "{cpu} " \
                  + "-m 2G " \
                  + "-nographic " \
                  + "{pflash} " \
                  + "-device virtio-net-pci,netdev=net0 " \
                  + "-netdev user,id=net0,hostfwd=tcp::{port}-:22" \
                  + "{ports} " \
                  + "-drive if=virtio,format=qcow2,file={img} " \
                  + "-drive if=virtio,format=raw,file={seed} " \
                  + "{custom} " \
                  + "> /dev/null; " \
                  + "rm {lock} {current}"

    _QEMU_ARM_START = ""
    LOCKFILE = QEMU_DIR + "qemu.lock"

    def _qemu_start(self):
        """
        start qemu machine
        """
        pwn.log.info(f"starting qemu machine, ssh port {self._port}")
        with open(Qegd.LOCKFILE, 'w') as lockfile:
            lockfile.write(str(self._port))
        pid = os.fork()
        if pid == 0:
            copyfile(Qegd.ARM_FLASH, Qegd.DEFAULT_QEMU_ARM_PFLASH)
            qemu_cmd = Qegd._QEMU_START.format(qemu=self._qemu,
                                               machine=" ".join((Qegd.DEFAULT_QEMU_MACHINE_PREFIX,
                                                                 self._machine)) if self._machine else '',
                                               cpu=" ".join(
                                                   (Qegd.DEFAULT_QEMU_CPU_PREFIX, self._cpu)) if self._cpu else '',
                                               pflash=" ".join((Qegd.DEFAULT_QEMU_PFLASH_PREFIX,
                                                                self._pflash)) if self._pflash else '',
                                               port=self._port,
                                               ports="".join(
                                                   [Qegd._QEMU_PORT_FORWARDING.format(host=host, guest=guest) for
                                                    host, guest in self._ports.items()]),
                                               img=Qegd.CURRENT_IMG,
                                               custom='',
                                               seed=Qegd.SEED_FILE,
                                               lock=Qegd.LOCKFILE,
                                               current=Qegd.CURRENT_IMG)
            pwn.log.info(qemu_cmd)
            os.system(qemu_cmd)
            sys.exit(0)

    def _new_vm(self) -> None:
        """
        create new vm
        """
        self._new = True

        pwn.log.info(f"no Lockfile in {Qegd.LOCKFILE}")
        self._set_local_img()
        self._setup_seed()
        # start qemu in independent process
        self._port = helper.first_free_port(Qegd.DEFAULT_PORT)
        self._qemu_start()
        time.sleep(15)

    def _vm_setup(self) -> None:
        """
        setup qemu machine
        """
        self._host = Qegd.DEFAULT_HOST
        if not os.path.exists(Qegd.LOCKFILE):
            self._new_vm()
        else:
            with open(Qegd.LOCKFILE, 'r') as lockfile:
                self._port = int(lockfile.readline())
            if not helper.is_port_in_use(self._port):
                pwn.log.info(f'Lockfile in {Qegd.LOCKFILE}, but no running machine detected. Creating new one')
                os.remove(Qegd.LOCKFILE)
                self._new_vm()
            pwn.log.info(f'Lockfile in {Qegd.LOCKFILE}. Using running qemu instance at port {self._port}')

    _TRIES = 3  # three times the charm

    def _ssh_setup(self) -> None:
        """
        setup ssh connection to QEMU
        """
        for _ in range(Qegd._TRIES):
            try:
                self._ssh = pwn.ssh(
                    user=self._user,
                    host=self._host,
                    port=self._port,
                    keyfile=Pwngd.KEYFILE,
                    ignore_config=True
                )
                break
            except:
                if _ + 1 == Qegd._TRIES:
                    pwn.log.error('SSH failed, pls try again')
                else:
                    pwn.log.info('Trying again')
                time.sleep(15)

    def __init__(self,
                 binary: str,
                 img: str = DEFAULT_IMG,
                 user: str = DEFAULT_USER,
                 ports: Dict[int, int] = None,
                 arm: bool = False,
                 qemu: str = DEFAULT_QEMU_CMD,
                 machine: str = DEFAULT_QEMU_MACHINE,
                 cpu: str = DEFAULT_QEMU_CPU,
                 pflash: str = None,
                 **kwargs):
        """

        :param binary: binary for VM debugging
        :param img: qemu image to use (requires ssh)
        :param user: user inside qemu image
        :param ports: forwarded ports
        :param arm: if qemu is arm
        :param qemu: qemu cmd
        :param cpu: value for :code -cpu
        :param machine: value for :code -machine
        :param pflash: value for :code -pflash
        :param kwargs: parameters to pass through to super
        """

        if not which('qemu-system-x86_64'):
            pwn.log.error('qemu-system-x86_64 isn\'t installed')

        if not os.path.exists(Qegd.QEMU_DIR):
            pwn.log.info(f"Generating {Qegd.QEMU_DIR} dir")
            os.makedirs(Qegd.QEMU_DIR)

        if arm:
            qemu = Qegd.DEFAULT_QEMU_ARM_CMD if qemu == Qegd.DEFAULT_QEMU_CMD else qemu
            cpu = Qegd.DEFAULT_QEMU_ARM_CPU if cpu == Qegd.DEFAULT_QEMU_CPU else cpu
            machine = Qegd.DEFAULT_QEMU_ARM_MACHINE if machine == Qegd.DEFAULT_QEMU_MACHINE else machine
            pflash = Qegd.DEFAULT_QEMU_ARM_PFLASH_OPTIONS + Qegd.DEFAULT_QEMU_ARM_PFLASH if pflash is None else pflash

        self._img = img
        self._user = user
        self._ports = ports if ports else dict()
        self._arm = arm
        self._qemu = qemu
        self._cpu = cpu
        self._machine = machine
        self._pflash = pflash

        self._vm_setup()
        self._ssh_setup()
        if self._new:
            self._install_packages(Pwngd.DEFAULT_PACKAGES)

        super().__init__(binary=binary, **kwargs)



class Box:
    """ constants class for recommended images and boxes """
    UBUNTU_JAMMY64 = 'ubuntu/jammy64'
    UBUNTU_FOCAL64 = 'ubuntu/focal64'
    UBUNTU_BIONIC64 = 'ubuntu/bionic64'
    UBUNTU_XENIAL64 = 'ubuntu/xenial64'
    CLOUDIMAGE_JAMMY = 'https://cloud-images.ubuntu.com/jammy/current/jammy-server-cloudimg-amd64.img'
    CLOUDIMAGE_FOCAL = 'https://cloud-images.ubuntu.com/focal/current/focal-server-cloudimg-amd64.img'
    CLOUDIMAGE_BIONIC = 'https://cloud-images.ubuntu.com/bionic/current/bionic-server-cloudimg-amd64.img'
    CLOUDIMAGE_XENIAL = 'https://cloud-images.ubuntu.com/xenial/current/xenial-server-cloudimg-amd64-disk1.img'
    CLOUDIMAGE_JAMMY_ARM = 'https://cloud-images.ubuntu.com/jammy/20230330/jammy-server-cloudimg-arm64.img'
    DOCKER_JAMMY = 'ubuntu:jammy'
    DOCKER_FOCAL = 'ubuntu:focal'
    DOCKER_BIONIC = 'ubuntu:bionic'
    DOCKER_XENIAL = 'ubuntu:xenial'



import os
if __name__ == '__main__':
    dir = os.path.dirname(os.path.realpath(__file__))
    with open(dir + '/template.txt', 'r') as template:
        for line in template.readlines():
            print(line, end='')
